@@ -1,7 +1,5 @@
-import collections
-
 from flex.decorators import rewrite_reserved_words
-from flex.utils import cast_value_to_type
+from flex.utils import dereference_reference
 
 
 @rewrite_reserved_words
@@ -32,20 +30,6 @@ def find_parameter(parameters, **kwargs):
     raise ValueError("No parameters matched")
 
 
-def type_cast_parameters(parameter_values, parameter_definitions):
-    typed_parameters = {}
-    for key in parameter_values.keys():
-        try:
-            parameter_definition = find_parameter(parameter_definitions, name=key)
-        except KeyError:
-            continue
-        if 'type' not in parameter_definition:
-            continue
-        value = parameter_values[key]
-        typed_parameters[key] = cast_value_to_type(value, parameter_definition['type'])
-    return typed_parameters
-
-
 def merge_parameter_lists(*parameter_definitions):
     """
     Merge multiple lists of parameters into a single list.  If there are any
@@ -59,8 +43,10 @@ def merge_parameter_lists(*parameter_definitions):
     return merged_parameters.values()
 
 
-def dereference_parameter_list(parameters, parameter_definitions):
-    return [
-        (p if isinstance(p, collections.Mapping) else parameter_definitions[p])
+def dereference_parameter_list(parameters, context):
+    return tuple((
+        dereference_reference(p['$ref'], context)
+        if '$ref' in p
+        else p
         for p in parameters
-    ]
+    ))
