@@ -5,16 +5,20 @@ import operator
 import functools
 import collections
 import itertools
+import json
 
 import six
 
-from django.core.validators import (
-    MinLengthValidator,
-    MaxLengthValidator,
+from flex.exceptions import (
+    ValidationError,
+    ErrorDict,
+    ErrorList,
+    MultiplePathsFound,
+)
+from flex.datastructures import (
+    ValidationDict,
+    ValidationList,
 )
-
-from flex.exceptions import ValidationError
-from flex.context_managers import ErrorCollection
 from flex.formats import registry
 from flex.utils import (
     is_value_of_any_type,
@@ -22,15 +26,16 @@ from flex.utils import (
     get_type_for_value,
     chain_reduce_partial,
     cast_value_to_type,
+    deep_equal,
 )
 from flex.paths import (
     match_path_to_api_path,
 )
 from flex.constants import (
-    EMPTY,
     NUMBER,
     STRING,
     ARRAY,
+    OBJECT,
     DELIMETERS,
     REQUEST_METHODS,
 )
@@ -38,13 +43,12 @@ from flex.decorators import (
     skip_if_not_of_type,
     skip_if_empty,
     suffix_reserved_words,
-    translate_validation_error,
 )
 from flex.error_messages import MESSAGES
 
 
 @skip_if_empty
-def validate_type(value, types):
+def validate_type(value, types, **kwargs):
     """
     Validate that the value is one of the provided primative types.
     """
@@ -85,21 +89,9 @@ def noop(*args, **kwargs):
     pass
 
 
-def validate_required(value):
-    if value is EMPTY:
-        raise ValidationError(MESSAGES['required']['required'])
-
-
-def generate_required_validator(required, **kwargs):
-    if required:
-        return validate_required
-    else:
-        return noop
-
-
 @skip_if_empty
 @skip_if_not_of_type(NUMBER)
-def validate_multiple_of(value, divisor):
+def validate_multiple_of(value, divisor, **kwargs):
     """
     Given a value and a divisor, validate that the value is divisible by the
     divisor.
@@ -116,7 +108,7 @@ def generate_multiple_of_validator(multipleOf, **kwargs):
 
 @skip_if_empty
 @skip_if_not_of_type(NUMBER)
-def validate_minimum(value, minimum, is_exclusive):
+def validate_minimum(value, minimum, is_exclusive, **kwargs):
     """
     Validator function for validating that a value does not violate it's
     minimum allowed value.  This validation can be inclusive, or exclusive of
@@ -144,7 +136,7 @@ def generate_minimum_validator(minimum, exclusiveMinimum=False, **kwargs):
 
 @skip_if_empty
 @skip_if_not_of_type(NUMBER)
-def validate_maximum(value, maximum, is_exclusive):
+def validate_maximum(value, maximum, is_exclusive, **kwargs):
     """
     Validator function for validating that a value does not violate it's
     maximum allowed value.  This validation can be inclusive, or exclusive of
@@ -170,27 +162,37 @@ def generate_maximum_validator(maximum, exclusiveMaximum=False, **kwargs):
     return functools.partial(validate_maximum, maximum=maximum, is_exclusive=exclusiveMaximum)
 
 
+@skip_if_empty
+@skip_if_not_of_type(STRING)
+def validate_min_length(value, minLength, **kwargs):
+    if len(value) < minLength:
+        raise ValidationError(MESSAGES['min_length']['invalid'].format(minLength))
+
+
 def generate_min_length_validator(minLength, **kwargs):
     """
     Generates a validator for enforcing the minLength of a string.
     """
-    return translate_validation_error(
-        skip_if_empty(skip_if_not_of_type(STRING)(MinLengthValidator(minLength).__call__)),
-    )
+    return functools.partial(validate_min_length, minLength=minLength)
+
+
+@skip_if_empty
+@skip_if_not_of_type(STRING)
+def validate_max_length(value, maxLength, **kwargs):
+    if len(value) > maxLength:
+        raise ValidationError(MESSAGES['max_length']['invalid'].format(maxLength))
 
 
 def generate_max_length_validator(maxLength, **kwargs):
     """
     Generates a validator for enforcing the maxLength of a string.
     """
-    return translate_validation_error(
-        skip_if_empty(skip_if_not_of_type(STRING)(MaxLengthValidator(maxLength).__call__)),
-    )
+    return functools.partial(validate_max_length, maxLength=maxLength)
 
 
 @skip_if_empty
 @skip_if_not_of_type(ARRAY)
-def validate_min_items(value, minimum):
+def validate_min_items(value, minimum, **kwargs):
     """
     Validator for ARRAY types to enforce a minimum number of items allowed for
     the ARRAY to be valid.
@@ -212,7 +214,7 @@ def generate_min_items_validator(minItems, **kwargs):
 
 @skip_if_empty
 @skip_if_not_of_type(ARRAY)
-def validate_max_items(value, maximum):
+def validate_max_items(value, maximum, **kwargs):
     """
     Validator for ARRAY types to enforce a maximum number of items allowed for
     the ARRAY to be valid.
@@ -234,16 +236,17 @@ def generate_max_items_validator(maxItems, **kwargs):
 
 @skip_if_empty
 @skip_if_not_of_type(ARRAY)
-def validate_unique_items(value):
+def validate_unique_items(value, **kwargs):
     """
     Validator for ARRAY types to enforce that all array items must be unique.
     """
     # we can't just look at the items themselves since 0 and False are treated
-    # the same as dictionary keys.
+    # the same as dictionary keys, and objects aren't hashable.
+
     counter = collections.Counter((
-        (v, type(v)) for v in value
+        json.dumps(v, sort_keys=True) for v in value
     ))
-    dupes = [v[0] for v, count in counter.items() if count > 1]
+    dupes = [json.loads(v) for v, count in counter.items() if count > 1]
     if dupes:
         raise ValidationError(
             MESSAGES['unique_items']['invalid'].format(
@@ -265,7 +268,7 @@ def generate_unique_items_validator(uniqueItems, **kwargs):
 
 @skip_if_empty
 @skip_if_not_of_type(STRING)
-def validate_pattern(value, regex):
+def validate_pattern(value, regex, **kwargs):
     if not regex.match(value):
         raise ValidationError(
             MESSAGES['pattern']['invalid'].format(value, regex.pattern),
@@ -276,19 +279,8 @@ def generate_pattern_validator(pattern, **kwargs):
     return functools.partial(validate_pattern, regex=re.compile(pattern))
 
 
-def deep_equal(a, b):
-    """
-    Because of things in python like:
-        >>> 1 == 1.0
-        True
-        >>> 1 == True
-        True
-    """
-    return a == b and isinstance(a, type(b)) and isinstance(b, type(a))
-
-
 @skip_if_empty
-def validate_enum(value, options):
+def validate_enum(value, options, **kwargs):
     if not any(deep_equal(value, option) for option in options):
         raise ValidationError(
             MESSAGES['enum']['invalid'].format(
@@ -301,21 +293,64 @@ def generate_enum_validator(enum, **kwargs):
     return functools.partial(validate_enum, options=enum)
 
 
-def validate_object(obj, validators):
+def validate_object(obj, field_validators=None, non_field_validators=None,
+                    schema=None, context=None):
     """
     Takes a mapping and applies a mapping of validator functions to it
     collecting and reraising any validation errors that occur.
     """
-    with ErrorCollection() as errors:
-        if '$ref' in validators:
-            ref_ = validators.pop('$ref')
-            for k, v in ref_.validators.items():
-                validators.setdefault(k, v)
-        for key, validator in validators.items():
+    if schema is None:
+        schema = {}
+    if context is None:
+        context = {}
+    if field_validators is None:
+        field_validators = ValidationDict()
+    if non_field_validators is None:
+        non_field_validators = ValidationList()
+
+    from flex.validation.schema import (
+        construct_schema_validators,
+    )
+    schema_validators = construct_schema_validators(schema, context)
+
+    if '$ref' in schema_validators and hasattr(schema_validators['$ref'], 'validators'):
+        ref_ = field_validators.pop('$ref')
+        for k, v in ref_.validators.items():
+            if k not in schema_validators:
+                schema_validators.add_validator(k, v)
+
+    schema_validators.update(field_validators)
+
+    schema_validators.validate_object(obj, context=context)
+    non_field_validators.validate_object(obj, context=context)
+
+    return obj
+
+
+def generate_object_validator(**kwargs):
+    return functools.partial(validate_object, **kwargs)
+
+
+@skip_if_empty
+@skip_if_not_of_type(OBJECT)
+def apply_validator_to_object(obj, validator, **kwargs):
+    with ErrorDict() as errors:
+        for key, value in obj.items():
             try:
-                validator(obj)
+                validator(value, **kwargs)
             except ValidationError as err:
-                errors[key].add_error(err.detail)
+                errors.add_error(key, err.detail)
+
+
+@skip_if_empty
+@skip_if_not_of_type(ARRAY)
+def apply_validator_to_array(values, validator, **kwargs):
+    with ErrorList() as errors:
+        for value in values:
+            try:
+                validator(value, **kwargs)
+            except ValidationError as err:
+                errors.add_error(err.detail)
 
 
 @suffix_reserved_words
@@ -369,7 +404,7 @@ def generate_value_processor(type_, collectionFormat=None, items=None, **kwargs)
                 functools.partial(cast_value_to_type, type_=type_)
             )
 
-    def processor(value):
+    def processor(value, **kwargs):
         try:
             return chain_reduce_partial(*processors)(value)
         except (ValueError, TypeError):
@@ -397,20 +432,22 @@ def validate_request_method_to_operation(request_method, path_definition):
     return operation_definition
 
 
-def validate_path_to_api_path(path, paths, basePath='', parameters=None, **kwargs):
+def validate_path_to_api_path(path, paths, basePath='', context=None, **kwargs):
     """
     Given a path, find the api_path it matches.
     """
-    if parameters is None:
-        parameters = {}
+    if context is None:
+        context = {}
     try:
         api_path = match_path_to_api_path(
             path_definitions=paths,
             target_path=path,
             base_path=basePath,
-            global_parameters=parameters,
+            context=context,
         )
-    except LookupError:
-        raise ValidationError(MESSAGES['path']['unknown_path'])
+    except LookupError as err:
+        raise ValidationError(str(err))
+    except MultiplePathsFound as err:
+        raise ValidationError(str(err))
 
     return api_path
