@@ -4,19 +4,25 @@ import functools
 
 import six
 
-from django.core.exceptions import ValidationError
-
-from flex.exceptions import SafeNestedValidationError
+from flex.exceptions import (
+    ValidationError,
+    ErrorList,
+    ErrorDict,
+)
+from flex.error_messages import MESSAGES
 from flex.constants import (
+    ARRAY,
     OBJECT,
-    EMPTY,
 )
 from flex.decorators import skip_if_not_of_type
+from flex.validation.reference import (
+    LazyReferenceValidator,
+)
 from flex.validation.common import (
+    noop,
     skip_if_empty,
     generate_type_validator,
     generate_format_validator,
-    generate_required_validator,
     generate_multiple_of_validator,
     generate_minimum_validator,
     generate_maximum_validator,
@@ -28,15 +34,38 @@ from flex.validation.common import (
     generate_pattern_validator,
     generate_enum_validator,
     validate_object,
+    generate_object_validator,
+)
+from flex.datastructures import (
+    ValidationDict,
 )
 
 
 @skip_if_empty
 @skip_if_not_of_type(OBJECT)
-def validate_min_properties(value, minimum):
+def validate_required(value, required_fields, **kwargs):
+    with ErrorDict() as errors:
+        for key in required_fields:
+            if key not in value:
+                errors.add_error(key, MESSAGES['required']['required'])
+
+
+def generate_required_validator(required, **kwargs):
+    if required:
+        return functools.partial(
+            validate_required,
+            required_fields=required,
+        )
+    else:
+        return noop
+
+
+@skip_if_empty
+@skip_if_not_of_type(OBJECT)
+def validate_min_properties(value, minimum, **kwargs):
     if len(value.keys()) < minimum:
         raise ValidationError(
-            "Object must have more than {0} properties.  It had {1}".format(
+            MESSAGES['min_properties']['invalid'].format(
                 minimum, len(value.keys()),
             ),
         )
@@ -48,10 +77,10 @@ def generate_min_properties_validator(minProperties, **kwargs):
 
 @skip_if_empty
 @skip_if_not_of_type(OBJECT)
-def validate_max_properties(value, maximum):
+def validate_max_properties(value, maximum, **kwargs):
     if len(value.keys()) > maximum:
         raise ValidationError(
-            "Object must have less than {0} properties.  It had {1}".format(
+            MESSAGES['max_properties']['invalid'].format(
                 maximum, len(value.keys()),
             ),
         )
@@ -64,32 +93,38 @@ def generate_max_properties_validator(maxProperties, **kwargs):
 def construct_items_validators(items, context):
     if isinstance(items, collections.Mapping):
         items_validators = construct_schema_validators(
-            items,
-            context,
+            schema=items,
+            context=context,
         )
     elif isinstance(items, six.string_types):
         items_validators = {
-            '$ref': LazyReferenceValidator(items, context),
+            '$ref': SchemaReferenceValidator(items, context),
         }
     else:
         assert 'Should not be possible'
     return items_validators
 
 
-def validate_items(objs, validators):
-    errors = []
-    for obj, validator in zip(objs, validators):
+@skip_if_not_of_type(ARRAY)
+@skip_if_empty
+def validate_items(objs, field_validators, **kwargs):
+    errors = ErrorList()
+    for obj, _field_validators in zip(objs, field_validators):
         try:
-            validate_object(obj, validator, inner=True)
+            validate_object(
+                obj,
+                field_validators=_field_validators,
+                **kwargs
+            )
         except ValidationError as e:
-            errors.extend(list(e.messages))
+            errors.add_error(e.detail)
 
     if errors:
-        raise SafeNestedValidationError(errors)
+        raise ValidationError(errors)
 
 
 def generate_items_validator(items, context, **kwargs):
-    if isinstance(items, collections.Mapping) or isinstance(items, six.string_types):
+    if isinstance(items, collections.Mapping):
         # If items is a reference or a schema, we pass it through as an
         # ever repeating list of the same validation dictionary, thus
         # validating all of the objects against the same schema.
@@ -110,7 +145,30 @@ def generate_items_validator(items, context, **kwargs):
     else:
         assert "Should not be possible"
     return functools.partial(
-        validate_items, validators=items_validators,
+        validate_items, field_validators=items_validators,
+    )
+
+
+@skip_if_not_of_type(OBJECT)
+@skip_if_empty
+def validate_additional_properties(obj, additional_properties, properties, **kwargs):
+    if additional_properties is False:
+        allowed_properties = set(properties.keys())
+        actual_properties = set(obj.keys())
+        extra_properties = actual_properties.difference(allowed_properties)
+        if extra_properties:
+            raise ValidationError(
+                MESSAGES['additional_properties']['extra_properties'].format(
+                    repr(extra_properties),
+                )
+            )
+
+
+def generate_additional_properties_validator(additionalProperties, properties, **kwargs):
+    return functools.partial(
+        validate_additional_properties,
+        additional_properties=additionalProperties,
+        properties=properties,
     )
 
 
@@ -134,41 +192,6 @@ validator_mapping = {
 }
 
 
-def validate_properties(obj, key, validators):
-    if obj is EMPTY:
-        return
-    validate_object(obj.get(key, EMPTY), validators, inner=True)
-
-
-class LazyReferenceValidator(object):
-    """
-    This class acts as a lazy validator for references in schemas to prevent an
-    infinite recursion error when a schema references itself, or there is a
-    reference loop between more than one schema.
-
-    The validator is only constructed if validator is needed.
-    """
-    def __init__(self, reference, context):
-        # TODO: something better than this assertion
-        assert 'definitions' in context
-        assert reference in context['definitions']
-        self.reference = reference
-        self.context = context
-
-    def __call__(self, value):
-        return validate_object(value, self.validators, inner=True)
-
-    @property
-    def validators(self):
-        return construct_schema_validators(
-            self.context['definitions'][self.reference],
-            self.context,
-        )
-
-    def items(self):
-        return self.validators.items()
-
-
 def construct_schema_validators(schema, context):
     """
     Given a schema object, construct a dictionary of validators needed to
@@ -185,29 +208,36 @@ def construct_schema_validators(schema, context):
             need recurse back into this function to generate a dictionary of
             validators for the property.
     """
-    validators = {}
+    validators = ValidationDict()
     if '$ref' in schema:
-        validators['$ref'] = LazyReferenceValidator(
-            schema['$ref'],
-            context,
+        validators.add_validator(
+            '$ref', SchemaReferenceValidator(schema['$ref'], context),
         )
     if 'properties' in schema:
-        # I don't know why this set intersection is enforced...?  Why did I do this.
-        intersection = set(schema['properties'].keys()).intersection(schema.keys())
-        assert not intersection
-
         for property_, property_schema in schema['properties'].items():
-            property_validators = construct_schema_validators(
-                property_schema,
-                context,
-            )
-            validators[property_] = functools.partial(
-                validate_properties,
-                key=property_,
-                validators=property_validators,
+            property_validator = generate_object_validator(
+                schema=property_schema,
+                context=context,
             )
+            validators.add_property_validator(property_, property_validator)
+    if schema.get('additionalProperties') is False:
+        validators.add_validator(
+            'additionalProperties',
+            generate_additional_properties_validator(context=context, **schema),
+        )
     assert 'context' not in schema
     for key in schema:
         if key in validator_mapping:
-            validators[key] = validator_mapping[key](context=context, **schema)
+            validators.add_validator(key, validator_mapping[key](context=context, **schema))
     return validators
+
+
+class SchemaReferenceValidator(LazyReferenceValidator):
+    """
+    This class acts as a lazy validator for references in schemas to prevent an
+    infinite recursion error when a schema references itself, or there is a
+    reference loop between more than one schema.
+
+    The validator is only constructed if validator is needed.
+    """
+    validators_constructor = construct_schema_validators
